import copy
import hashlib
import logging
import numpy as np
import simplejson as json

from os.path import join

from cvdatasets.annotation.base import BaseAnnotations
from cvdatasets.annotation.mixins.bbox_mixin import BBoxMixin
from cvdatasets.annotation.mixins.parts_mixin import PartsMixin
from cvdatasets.utils import _MetaInfo


def _uuid_entry(im_info):
	return hashlib.md5(im_info["file_name"].encode()).hexdigest()

class BaseINAT_Annotations(BBoxMixin, PartsMixin, BaseAnnotations):

	def read_content(self, json_file, attr, optional=False):
		if not json_file.endswith(".json"):
			return super(BaseINAT_Annotations, self).read_content(json_file, attr, optional=optional)

		self.set_content_from_file(json_file, attr, json.load, optional)

	def _load_bounding_boxes(self):
		self.bounding_boxes = np.zeros(len(self.uuids), dtype=self.meta.bounding_box_dtype)

		for i, im in enumerate(self._content["images"]):
			self.bounding_boxes[i]["w"] = im["width"]
			self.bounding_boxes[i]["h"] = im["height"]

	def _load_split(self):
		self.train_split = np.ones(len(self.uuids), dtype=bool)
		val_uuids = [str(im["id"]) for im in self._val_content["images"]]
		for v_uuid in val_uuids:
			self.train_split[self.uuid_to_idx[v_uuid]] = False

		self.test_split = np.logical_not(self.train_split)

	def _load_labels(self):
		self.labels = np.zeros(len(self.uuids), dtype=np.int32)
		labs = {str(annot["image_id"]): annot["category_id"] for annot in self._content["annotations"]}
		for uuid in self.uuids:
			self.labels[self.uuid_to_idx[uuid]] = labs[uuid]

	def _load_uuids(self):
		uuid_fnames = [(str(im["id"]), im["file_name"]) for im in self._content["images"]]
		self.uuids, self.images = map(np.array, zip(*uuid_fnames))

		assert len(np.unique(self.uuids)) == len(self.uuids), \
			"UUIDs are not unique!"

		self.uuid_to_idx = {uuid: i for i, uuid in enumerate(self.uuids)}

	@property
	def meta(self):
		raise NotImplementedError

class INAT20_Annotations(BaseINAT_Annotations):

	name="INAT20"

	@property
	def meta(self):
		info = _MetaInfo(
			images_folder="images",
			content="trainval.json",
			val_content="val.json",
			unlabeled_content="unlabeled_train.json",

			# fake bounding boxes: the whole image
			bounding_box_dtype=np.dtype([(v, np.int32) for v in "xywh"]),

			parts_file=join("parts", "part_locs.txt"),
			part_names_file=join("parts", "parts.txt"),
		)

		info.structure = [
			[info.content, "_content"],
			[info.val_content, "_val_content"],
			[info.unlabeled_content, "_unlabeled_content", True],
			[info.parts_file, "_part_locs", True],
			[info.part_names_file, "_part_names", True],
		]
		return info

	@property
	def has_unlabeled_data(self):
		return self._unlabeled_content is not None

	def _load_uuids(self, *args, **kwargs):
		super(INAT20_Annotations, self)._load_uuids(*args, **kwargs)

		if not self.has_unlabeled_data:
			logging.info("No unlabeled data was provided!")
			return

		logging.info("Loading unlabeled data...")
		uuid_fnames = [(_uuid_entry(im), im["file_name"]) for im in self._unlabeled_content["images"]]

		self.unlabeled = unlabeled = copy.copy(self)

		unlabeled.uuids, unlabeled.images = map(np.array, zip(*uuid_fnames))
		unlabeled.labels = np.full(unlabeled.images.shape, -1, dtype=np.int32)
		unlabeled.train_split = np.full(unlabeled.images.shape, 1, dtype=bool)
		unlabeled.test_split = np.full(unlabeled.images.shape, 0, dtype=bool)

		assert len(np.unique(unlabeled.uuids)) == len(unlabeled.uuids), \
			"Unlabeled UUIDs are not unique!"

		overlap = set(self.uuids) & set(unlabeled.uuids)
		assert len(overlap) == 0, \
			f"Unlabeled and labeled UUIDs overlap: {overlap}"

		unlabeled.uuid_to_idx = {uuid: i for i, uuid in enumerate(unlabeled.uuids)}



class INAT19_Annotations(BaseINAT_Annotations):

	name="INAT19"

	@property
	def meta(self):
		info = _MetaInfo(
			images_folder="images",
			content="trainval2019.json",
			val_content="val2019.json",
			# train_content="train2019.json",

			# fake bounding boxes: the whole image
			bounding_box_dtype=np.dtype([(v, np.int32) for v in "xywh"]),

			parts_file=join("parts", "part_locs.txt"),
			part_names_file=join("parts", "parts.txt"),
		)

		info.structure = [
			[info.content, "_content"],
			[info.val_content, "_val_content"],
			[info.parts_file, "_part_locs"],
			[info.part_names_file, "_part_names"],
		]
		return info

class INAT18_Annotations(BaseINAT_Annotations):

	name="INAT18"

	@property
	def meta(self):
		info = _MetaInfo(
			images_folder="images",
			content="trainval2018.json",
			# content="train2018.json",
			val_content="val2018.json",

			# fake bounding boxes: the whole image
			bounding_box_dtype=np.dtype([(v, np.int32) for v in "xywh"]),

			parts_file=join("parts", "part_locs.txt"),
			part_names_file=join("parts", "parts.txt"),
		)

		info.structure = [
			[info.content, "_content"],
			[info.val_content, "_val_content"],
			[info.parts_file, "_part_locs"],
			[info.part_names_file, "_part_names"],
		]
		return info
